################################################################################
## Utility functions for OPUS which use pyFits
##
## Methods:
##     get_asn_members(filename)
##     makeSPTcopy(dataset, inSuffix, outSuffix, extension )
## 
################################################################################

import os
import pyfits
import opusutil

################################################################################
##
## Method: get_asn_members
## 
## Description:
##    We are after is a list of existing members
##    The MEMTYPE column should contain 'EXP', and 
##        MEMPRSNT should be "yes"
##    Return a list of 'MEMNAME's
##    Not that we ever would, but this method should also work for a
##         NICMOS ASC table array, even though it has extra columns.
##         That is the beauty of pyFITS.
##
## Input:
##      filename - name of file
##                 expecting FILETYPE = 'ASN_TABLE', else errors
##
## Return:
##   members - an array of present non-PRODuct members, i.e. members with 
##             MEMTYPE  like 'EXP%'
##             MEMPRSNT like yes, a boolean, how exciting
##             Expect these member names to be capitalized
##  []       - empty array should be checked for and considered an error
##
################################################################################
def get_asn_members(filename):
    methodName = 'get_asn_members' 

    if not os.path.exists(filename):
        msg = "FATAL: file does NOT exist: " + filename
        opusutil.PrintMsg('E', msg, methodName )
        return 1

    hdulist = pyfits.open(filename)
    prihdr = hdulist[0].header
    ## Verify this is indeed an ASN_TABLE
    if prihdr['FILETYPE'] != 'ASN_TABLE':
        msg = "NOT an ASN table"
        opusutil.PrintMsg("E", msg, methodName )
        ## Callers should check that returned array is not empty
        return []

    ## We know ASN table is in first extension
    tbldata = hdulist[1].data

    members = []
    ## These assignments like an alias to a shorter name   
    ## Since I do not expect to change the arrays, do NOT need copy.deepcopy
    memtypes = tbldata.field('MEMTYPE') 
    memnames = tbldata.field('MEMNAME') 
    memprsnt = tbldata.field('MEMPRSNT')
    ## We know the list of members in in tbldata.field('MEMNAME')
    ## BUT we only want those whose tbldata.field('MEMTYPE') matches 'EXP'
    ## and those who are present (boolean yes)
    for a_cnt in xrange( len(memtypes) ):
        memtype = memtypes[a_cnt]
        member  = memnames[a_cnt]
        present = memprsnt[a_cnt]
        if memtype.find('EXP') >= 0 and present:
            msg =  "appending member " + member + " to members"
            opusutil.PrintMsg ("D", msg )
            members.append(member) 
   
    return members 



##------------------------------------------------------------------------------
##
## Method: makeSPTcopy
##
## Description: 
##     Use pyfits to copy the correct COS SPT extensions to a new file 
##         for cataloging
##     Do not need to make this more generic because 
##         rawacq files can just be file-copied
##
##     Extensions can be found by pyFITS in two ways: by name or by number.
##     We expect to have multiple pairs of SNAP1 and SNAP2 extensions, therefore
##         we need to use the number.
##
##     This is how I decided which SPT extension to copy:
##         For SNAP2, probably the last one: NEXTEND would be the number
##         For SNAP1, still waiting to hear if it is 
##                Last pair's SNAP1 [NEXTEND - 1], or
##                first SNAP1 in file extension = [2].
##         I'm going to guess Last pair's SNAP1 until I hear otherwise.
##         I believe we heard otherwise, and now 2 is hard-coded. 12/30/08
##
## Inputs:
##    dataset   - rootname to copy 
##    inSuffix  - suffix for source file
##    outSuffix - suffix for destination file 
##    extension - which extension to put in the destination file
##
## Returns: (creates a new file)
##    0 -- success
##    1 -- failure -- How do I detect a failure in the pyfits?
##
##------------------------------------------------------------------------------
def makeSPTcopy(dataset, inSuffix, outSuffix, extension ):
    from ac_pkg import rmOldFile

    methodName = "makeSPTcopy"   ## For messages

    oldName = dataset + "_" + inSuffix + ".fits"
    newName = dataset + "_" + outSuffix + ".fits"

    rmOldFile(newName, methodName)

    ## Which do we like better?  1st: reads better; 2nd: ext easy to find
    msg = "Copy " + oldName +" "+ extension  + " to " + newName 
#   msg = "Copy " + oldName + " to " + newName + " for " + extension  
    opusutil.PrintMsg('I', msg)
    msg  = "Use pyFITS to copy " + extension + " from " 
    msg += oldName + " to " + newName
    opusutil.PrintMsg("D", msg)

    if not os.path.exists(oldName):
        msg = "Filename " + oldName + "does NOT exist"
        opusutil.PrintMsg('F', msg, methodName)
        return 1

    ## Read in source file
    SPT = pyfits.open( oldName )
    SPT.close()
    if extension == "SNAP2":
        ## NEXTEND should correspond to the last SNAP2 header
        ext = SPT[0].header['NEXTEND']
    elif extension == "SNAP1":
        ## Last pair's SNAP1
       #ext = SPT[0].header['NEXTEND'] - 1
        ## If I need the first SNAP1, then it will be "2" hard-coded
        ext = 2
    msg = "Copying extension " + str(ext) + ": " + SPT[ext].header['EXTNAME']
    opusutil.PrintMsg('D', msg)

    ## pyFITS magic
    newSPT = pyfits.HDUList()
    newSPT.append(SPT['PRIMARY'])
    newSPT.append(SPT[ext])
    ## Update NEXTEND for the new file
    ## (Have to subtract 1 because PRIMARY doesn't count)
    newSPT[0].header['NEXTEND'] = len(newSPT) - 1
    try:
        newSPT.writeto(newName)
    except:
        msg = "Failed to create " + newName
        opusutil.PrintMsg('F', msg, methodName)
        return 1
    newSPT.close()
    return 0                    ## Success
    

##------------------------------------------------------------------------------
##
## Method: update_qual_keywords
##
## Description: 
##     Use pyfits to update or add the QUALCOM[1-3] and QUALITY keywords of 
##         headers in provided list of files
##
## Inputs:
##    file_list
##         - a list of files in current dir whose headers need to be updated
##           with the same QUAL* information 
##    pdq_summary_results
##         - a list of dicts containing pdq_summary table information
##           (but truly I am only expecting 1 dict)
##
## Returns: (updates header keywords)
##    0 -- success
##    RuntimeError -- failure -- How do I detect a failure in the pyfits?
##         only by the exception raised?  where put the try?  what do if fail?
##
##   I am not sure I approve of trying to capture these errors if 
##   one looses valueable information about it when one does not 
##   know what exception to expect? ???
##
## History:
## mm/dd/yy PR    Who        What          
## -------- ----- ---------- --------------------------------------------------
## 06/03/10 64909 Sherbert   update singletons, products, subproducts
##
##------------------------------------------------------------------------------
def update_qual_keywords( file_list, pdq_summary_results ):
    nm = 'update_qual_keywords'
    msg = "input file_list          : " + str(file_list)
    opusutil.PrintMsg( 'D', msg, nm )
    msg = "input pdq_summary_results: " + str(pdq_summary_results)
    opusutil.PrintMsg( 'D', msg, nm )

    ##            keyVal    : keyHash[keyVal]
    keyHash   = {'QUALCOM1' : 'pdq_comment_1' ,
                 'QUALCOM2' : 'pdq_comment_2' , 
                 'QUALCOM3' : 'pdq_comment_3' , 
                 'QUALITY'  : 'pdq_quality'  
                }
    

    ## pdq_summary_results is a list containing 1 dict (pdq_summary_results[0])
    pdqSumRow = pdq_summary_results[0]

    ## Update keywords in each file
    for file in file_list:
       #msg = 'update_qual_keywords: about to update' + file
       #opusutil.PrintMsg( 'D', msg, nm )
        try:
            hdulist = pyfits.open(file, mode='update')
            prihdr = hdulist[0].header
        except:
            errMsg = "Either could not open " + file 
            errMsg += " or could not read primary header "
            raise RuntimeError, errMsg

        for keyVal in keyHash.keys():
            msg  = "Set " + file + " keyword " + keyVal + " = "
            msg += str(pdqSumRow[ keyHash[keyVal] ])
            opusutil.PrintMsg( 'D', msg, nm )
            ## We want these keywords to update in headers when they exist, 
            ## but NOT be added to headers from which they are missing.  
            ## Exception thrown if attempt to update header w/o QUAL* keywords.
            if ( pdqSumRow[ keyHash[keyVal] ] != None ):
                try:
                    prihdr[keyVal] = pdqSumRow[ keyHash[keyVal] ]
                except:
                    errMsg  = "Could not update " + keyVal 
                    errMsg += " in file " + file 
                    errMsg += ".  Perhaps keyword does not exist in file?"
                    raise RuntimeError, errMsg
        hdulist.close()

    return 0




##------------------------------------------------------------------------------
##
## Method: update_coord_keywords
##
## Description: 
##     Use pyfits to update or add the [GE]{LAT|LON}_REF keywords of 
##         primary headers in provided list of files 
##
##     Use the opuscoords package in stsci_python which Warren Hack kindly
##     created from gc_coords_pkg.c methods GC_radec2glatlon  and 
##     GC_radec2elatlon. 
##
## Inputs:
##    file_list
##         - a list of files in current dir whose headers need to be updated
##           These files should have only one SCI extension, eg. NIC mos files
##
## Returns: (updates header keywords)
##    0 -- success
##    RuntimeError -- failure -- How do I detect a failure in the pyfits?
##         only by the exception raised?  where put the try?  what do if fail?
##
##   I am not sure I approve of trying to capture these errors if 
##   one looses valueable information about it when one does not 
##   know what exception to expect? ???
##
## History:
## mm/dd/yy PR    Who        What
## -------- ----- ---------- ---------------------------------------------------
## 01/20/09 61586 Sherbert/Hack Recompute NIC coords after calib
## 
##------------------------------------------------------------------------------
def update_coord_keywords( file_list ):

    import opuscoords
   
    thisModule = 'update_coord_keywords'
    for file in file_list:
        msg  = "File being updated is " + file 
        opusutil.PrintMsg('D', msg, thisModule )
        try:
            hdulist = pyfits.open(file, mode='update')
            prihdr = hdulist[0].header
            scihdr = hdulist[1].header
        except:
            errMsg = "Either could not open " + file 
            errMsg += " or could not read primary header "
            raise RuntimeError, errMsg

        ## Get the input from the [SCI,1] headers
        try:
            ra  = scihdr['CRVAL1']
            dec = scihdr['CRVAL2']
        except:
            errMsg  = "One or both keywords CRVAL[12] not found in header: "
            errMsg += scihdr
            raise RuntimeError, errMsg

        ## TRACEs
        msg = "Input RA  value is " + str( ra ) 
        opusutil.PrintMsg('D', msg, thisModule )
        msg = "Input DEC value is " + str( dec ) 
        opusutil.PrintMsg('D', msg, thisModule )
   
        ## Calculate the new values
        glon, glat = opuscoords.radec2glatlon( ra, dec )
        ## Update the keywords in the headers
        try:
            prihdr.update( 'GLAT_REF', glat )
            prihdr.update( 'GLON_REF', glon )
        except:
            errMsg  = "One/both keywords G{LAT|LON}_REF not found in header: "
            errMsg += prihdr
            raise RuntimeError, errMsg

        ## TRACEs
        msg = "New GLAT_REF value is " + str( glat ) 
        opusutil.PrintMsg('D', msg, thisModule )
        msg = "New GLON_REF value is " + str( glon ) 
        opusutil.PrintMsg('D', msg, thisModule )
   
        ## Calculate the new values
        elon, elat = opuscoords.radec2elatlon( ra, dec )
        ## Update the keywords in the headers
        try:
            prihdr.update( 'ELAT_REF', elat )
            prihdr.update( 'ELON_REF', elon )
        except:
            errMsg  = "One/both keywords E{LAT|LON}_REF not found in header: "
            errMsg += prihdr
            raise RuntimeError, errMsg

        ## TRACEs
        msg = "New ELAT_REF value is " + str( elat ) 
        opusutil.PrintMsg('D', msg, thisModule )
        msg = "New ELON_REF value is " + str( elon ) 
        opusutil.PrintMsg('D', msg, thisModule )
   
        hdulist.close()
   
    return 0

